/***************************** C SOURCE FILE *********************************
**
**  Project:    Logic IO RTCU Gateway Protocol
**  Filename:   GWPROT.C 
**  
******************************************************************************
**
*****************************************************************************/


/*****************************************************************************
**
** MODULES USED 
**
*****************************************************************************/
#define GW_DEBUG              0
#include <windows.h>
#include <winsock.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <conio.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include "..\Modules\queue.h"
#include "..\Modules\gwcrc16.h"
#include "..\Modules\compress.h"
#include "..\Modules\rijndael128.h"
#include "..\Modules\security.h"
#include "gwprot.h"
#include "racp2.h"
#include "gwprotld.h"


/*****************************************************************************
**
**    DEFINITIONS AND MACROS
**
*****************************************************************************/
#if GW_DEBUG
  #define debug(x)                     debug_msg x
#else            
  #define debug(x)
#endif

/* Following two constants are used for realize throttling when the gateway can not be reached.
 * The throttling will delay the disconnection of the GPRS connection with increasing steps of
 * GW_CON_DELAY_STEP seconds until GW_CON_MAX_DELAY has been reached. When this happens the
 * throttling will start from the beginning again.
 *
 * For precise timing GW_CON_MAX_DELAY MOD GW_CON_DELAY_STEP must be 0 (zero).
 */
#define GW_CON_DELAY_STEP                30  /* time increase in seconds for each throttle step */
#define GW_CON_MAX_DELAY                300  /* maximum delay in seconds for internet closedown */

#define GWPROT_MAX_CLIENTS             1000          /* maximum number of clients */
#define GWPROT_MAX_PACKETS_STD         7000          /* maximum number of packets in circulation */
#define GWPROT_MAX_PACKETS_EXT         1000          /* maximum number of packets in circulation */

#define THROTTLE_ENABLE                   0

#define CREATETHREAD(th,arg)           _beginthread(th,0x20000,(void*)arg);
#define SLEEP(s)                       Sleep(s)
#define SEMAPHORE                      HANDLE
#define SEM_CREATE(s,v)                s=CreateSemaphore(NULL,v,0xffff,NULL)
#define SEM_SIGNAL(s)                  ReleaseSemaphore(s,1,NULL)
#define SEM_WAIT(s)                    WaitForSingleObject(s,INFINITE)
#define SEM_WAIT_T(s,t)                (WaitForSingleObject(s,t)!=WAIT_OBJECT_0)
#define SEM_CLOSE(s)                   CloseHandle(s)
#define INIT_CRIT                      InitializeCriticalSection(&gw_critical_section)
#define ENTER_CRIT                     EnterCriticalSection(&gw_critical_section)
#define EXIT_CRIT                      LeaveCriticalSection(&gw_critical_section)
#define GET_RANDOM()                   ((unsigned long)(rand()<<16 | rand()+rand()))
#define INIT_RANDOM                    srand((unsigned) time(0))

#define CLIENT_CACHE_SIZE                 3    /* number of entries cached for each client */
#define TIMER_TICK_FRQ                 5000
#define CACHE_ALIVE_TICKS                60    /* milliseconds=CACHE_ALIVE_TICKS*TIMER_TICK_FRQ */
#define GW_MAXCHANNELS                    8

#define MAX_CONNECTION_RETRIES         (pSession->max_connection_attempt)
#define MAX_SEND_REQ_RETRIES           (pSession->max_send_req_attempt)
#define RESPONSE_TIMEOUT               (pSession->response_timeout)
#define ACTIVITY_MONITOR_THRESHOLD     (pSession->alive_freq)

#define DATALENGTH(item)               ((item)->type==2 ? ((struct ext_packet*)(item))->hdr.data_len : (item)->hdr.data_len)
#define HEADLENGTH(item)               ((item)->type==2 ? sizeof(struct ext_header) : sizeof(struct racp2_header) )
#define FRAMEDATA(item)                ((item)->type==2 ? ((struct ext_packet*)(item))->data : (item)->data)
#define MAXLENGTH(item)                ((item)->type==2 ? GW_MAXDATASIZE_EXT : GW_MAXDATASIZE_STD)
#define CHECKEXT(item)                 (((struct ext_packet*)(item))->hdr.extpack==3)

#define SET_LENGTH(item, val)          { if ((item)->type==2) ((struct ext_packet*)(item))->hdr.data_len=(val); else (item)->hdr.data_len=(val); }


/****************************************************************************
**
**    TYPEDEFS AND STRUCTURES
**
****************************************************************************/
struct client_info {
   qnode                      qlink;
   unsigned long              node_id;
   unsigned long              last_used;        //timestamp for tail of cacheq last-used.
   queue                      cacheq;           //queue of cached responses. (racp2_packet)
};

struct request_info {
   qnode                      qlink;
   unsigned long              tno;
   unsigned char              gw_response;
   void*                      data;
   unsigned short             data_len;
   unsigned short             rc;
   SEMAPHORE                  rxsignal;
};

struct hSession {
   unsigned char              opened;                               //is the gateway session open
   unsigned char              initialized;                          //has the session been initialized?
   unsigned long              our_node_id;                          //our node id;
   unsigned long              initial_node_id;
   char                       gw_address[42];                       //gateway address
   unsigned short             gw_port;                              //gateway port
   char                       gw_key[10];                           //gateway key.
   unsigned char              gw_connection;                        //is connection to the gateway established?
   unsigned char              gw_logon_in_progress;
   unsigned char              userTransactionsOK;                   //has user transaction been succesfully done?
   unsigned char              forceSelfTransaction;                 //will force a selftrans. through now.
   unsigned char              running_threads;                      //number of running service threads
   char                       tls_server[512];                      //path to the file used to validate the server certificate
   char                       tls_cert[512];                        //path to the file with the certificate to identify gwprot
   char                       tls_key[512];                         //path to the file with the private key of the certificate
   char                       tls_pass[51];                         //The password for the private key

   queue                      client_list;                          //client list (client_info);
   queue                      client_freelist;                      //client freelist (client_info)
   struct client_info         client_buffers[GWPROT_MAX_CLIENTS];   //the raw client_info descriptors.

   queue                      std_freelist;                         //racp2 buffer free list
   struct racp2_packet        std_buffers[GWPROT_MAX_PACKETS_STD];  //the raw racp2 buffers
   SEMAPHORE                  std_freelist_notempty;                //event semaphore for freelist
   queue                      ext_freelist;                         //racp2 buffer free list
   struct ext_packet          ext_buffers[GWPROT_MAX_PACKETS_EXT];  //the raw racp2 buffers
   SEMAPHORE                  ext_freelist_notempty;                //event semaphore for freelist

   queue                      out_requests;                         //queue of (request_info) outstanding requests.
   queue                      in_requests[GW_MAXCHANNELS];          //queue of (racp2_packet) incoming requests for each channel.
   SEMAPHORE                  in_requests_notempty[GW_MAXCHANNELS]; //event semaphore for in_requests.

   unsigned long              next_tx_tno;                          //next TX TNO
   SEMAPHORE                  send_mutex;                           //mutex for performing send request.

   unsigned long              time_tick;

   SEMAPHORE*                 logon_rxsignal;                       //used for cancelling during logon
   unsigned char              logon_cancelled;                      //used for cancelling during logon

   char                       comp_buf[GW_MAXDATASIZE_EXT+8];       // Compression and Encryption buffer
   SEMAPHORE                  comp_mutex;                           // Thread guard for Compression and Encryption buffer
   unsigned char              comp_enable;                          // is Compression enabled.
   unsigned char              gw_apply_compress;                    // Apply compression to transactions
   unsigned char              gw_apply_encrypt;                     // Apply encryption to transactions
   unsigned short             gw_extended_maxsize;                  // Maximum extended frame size allowed by gateway
   unsigned short             ext_maxsize;                          // Maximum extended frame size specified
   unsigned char              crypt_key[16];                        // Key used to encrypt packages to gateway
   unsigned char              crypt_key_set;                        // is encryption key specified.

   struct racp2_packet        req_packet_std;
   struct ext_packet          req_packet_ext;

   //configurable parameters:
   unsigned char              max_connection_attempt;
   unsigned char              max_send_req_attempt;
   unsigned short             response_timeout;
   unsigned short             alive_freq;

   struct sock_inst_t*        conn;
};


/*****************************************************************************
**
** EXPORTED VARIABLES 
**
*****************************************************************************/


/*****************************************************************************
**
** GLOBAL VARIABLES 
**
*****************************************************************************/
static unsigned char          initialized = 0;  //has the module been initialized?
static struct hSession*       gwSession = NULL;  //a single gateway session
static CRITICAL_SECTION       gw_critical_section;

static const char             GW_VER_REQUEST[9] = { 0xDA, 0xFE, 0x11, 0x16, 0x69, 0xEF, 0x08, 0x06, 0x00 };


/*****************************************************************************
**
** LOCAL FUNCTIONS 
**
*****************************************************************************/

#if GW_DEBUG
/////////////////////////////////////////////////////////////////////////////
// Write debug information
//
// Input:
//       format
//       args...
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
void debug_msg(char* fmt, ...)
{
   static char buf[300];
   va_list argp;

   ENTER_CRIT;
   va_start(argp, fmt);
   vsprintf(buf, fmt, argp);
   va_end(argp);
   cprintf(buf);
   EXIT_CRIT;
}
#endif

/////////////////////////////////////////////////////////////////////////////
// Release an allocated racp2 packet.
//
// Input:
//      pSession   - Pointer to the session that wants to allocate the racp2
//                   packet.
//      type       - The type of frame to allocate. 1=Standard, 2=Extended
//
// Output:
//      None.
//
// Returns:
//      Pointer to racp2 packet.
//      NULL       - No racp2 packets could be allocated.
//
/////////////////////////////////////////////////////////////////////////////
static struct racp2_packet* allocPacket(struct hSession* pSession, int type)
{
   // Var
   struct racp2_packet* rp;
   SEMAPHORE*           notempty;
   queue*               freelist;

   // Extended frame
   if (type==2) {
      notempty = &pSession->ext_freelist_notempty;
      freelist = &pSession->ext_freelist;
   }
   // Standard frame
   else {
      notempty = &pSession->std_freelist_notempty;
      freelist = &pSession->std_freelist;
   }

   // Wait for free packet
   if (SEM_WAIT_T(*notempty, 100)) return NULL;

   // Get packet from queue
   ENTER_CRIT;
   rp = (struct racp2_packet*)qRemoveHead(freelist);
   EXIT_CRIT;

   // Clear packet flags
   if (rp) {
      rp->hdr.encrypt     = 0;
      rp->hdr.compress    = 0;
      rp->hdr.gw_response = 0;
      if (rp->type==2) {
         ((struct ext_packet*)rp)->hdr.priority  = 0;
         ((struct ext_packet*)rp)->hdr.unack     = 0;
         ((struct ext_packet*)rp)->hdr.reserved1 = 0;
         ((struct ext_packet*)rp)->hdr.extpack   = 3;
         ((struct ext_packet*)rp)->hdr.reserved2 = 0;
      }
   }
   return rp;
}

/////////////////////////////////////////////////////////////////////////////
// Release an allocated racp2 packet.
//
// Input:
//      pSession   - Pointer to the session.
//      rc         - racp2 packet to release.
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
static void freePacket(struct hSession *pSession, struct racp2_packet* rp)
{
   // Var
   SEMAPHORE* notempty;
   queue*     freelist;

   // Extended frame
   if (rp->type==2) {
      notempty = &pSession->ext_freelist_notempty;
      freelist = &pSession->ext_freelist;
   }
   // Standard frame
   else {
      notempty = &pSession->std_freelist_notempty;
      freelist = &pSession->std_freelist;
   }

   ENTER_CRIT;
   qInsertHead(freelist, (qnode*)rp);
   EXIT_CRIT;
   SEM_SIGNAL(*notempty);
}

/////////////////////////////////////////////////////////////////////////////
// Validate CRC in a received racp2 packet.
//
// Input:
//      rc       - racp2 packet to release.
//
// Output:
//      None.
//
// Returns:
//      0        - CRC ok.
//      1        - Invalid data length in racp2 packet.
//      2        - CRC is not correct.
//
/////////////////////////////////////////////////////////////////////////////
static unsigned short checkCRC(struct racp2_packet* rp)
{
   // Var
   unsigned short i_crc, c_crc;
   unsigned short datalen, headlen, maxsize;

   // Get size
   datalen = DATALENGTH(rp);
   headlen = HEADLENGTH(rp);
   maxsize = MAXLENGTH(rp);

   // Validate
   if (datalen > maxsize) return 1;
   if (rp->hdr.crc == 0) return 0;    //crc==0 is always accepted..

   // Calculate CRC
   i_crc       = rp->hdr.crc;
   rp->hdr.crc = 0;
   c_crc       = gw_calc_crc16data(0, (unsigned char*)&rp->hdr, headlen + datalen);
   rp->hdr.crc = i_crc;

   // Completed
   if (c_crc != i_crc) {
      debug(("checkCRC. TNO=%lu, calc=0x%x, packet=0x%x\n", rp->hdr.TNO, c_crc, i_crc));
      return 2;
   }
   return 0;
}

/////////////////////////////////////////////////////////////////////////////
// Handles the response cache.
//
// Input:
//      pSession   - Pointer to the session.
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
static void ageResponseCache(struct hSession *pSession)
{
   // Var
   struct client_info* ci;
   struct client_info* ri;
   unsigned long       now = pSession->time_tick;
   int                 n = 0;

   // Lock
   ENTER_CRIT;

   // Iterate clients
   ci = (struct client_info*)qPeekHead(&pSession->client_list);
   while (ci) {
      if (qSize(&ci->cacheq)) {
         //something in the cache, so the timestamp can be used:
         if (now - ci->last_used > CACHE_ALIVE_TICKS) {
            //remove the oldest record and free it.
            freePacket(pSession, (struct racp2_packet*)qSelRemove(&ci->cacheq, qPeekTail(&ci->cacheq)));

            //mark new timestamp:
            ci->last_used = now;
            ri            = ci;
            n++;
            break;
         }
      }
      ci = (struct client_info*)(ci->qlink.next);
   }

   // Unlock
   EXIT_CRIT;

   // Completed
   if (n) {
      debug(("[0x%p]ageResponseCache removed %d from client nodeid=%lu\n", pSession, n, ri->node_id));
   }
}

/////////////////////////////////////////////////////////////////////////////
// Query this nodes node_id.
// This function is used especially when a dynamic node-id is requested.
//
// Input:
//      pSession   - Pointer to the session.
//      buf        - Pointer to buffer where received data is stored.
//      size       - The number of bytes to receive.
//
// Output:
//      None.
//
// Returns:
//      0          - Success.
//      1          - Failure.
//
/////////////////////////////////////////////////////////////////////////////
static unsigned short blockReceive(struct hSession* pSession,
                                   void*            buf,
                                   unsigned short   size)
{
   // Var
   unsigned short write_pos, rd_len;

   // Do until data block is received
   for (write_pos = 0; write_pos < size; write_pos += rd_len) {
      if (gwldReceive(pSession->conn, size - write_pos, &((char*)buf)[write_pos], &rd_len)) {
         //failed
         debug(("[0x%p]blockReceive: failed..\n", pSession));
         return 1;
      }
   }

   // Completed
   return 0;
}

/////////////////////////////////////////////////////////////////////////////
// Handles a received racp2 packet as a responce.
//
// Input:
//      pSession   - Pointer to the session.
//      rp         - The racp2 packet received.
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
static void handleResponse(struct hSession *pSession,
                           struct racp2_packet* rp)
{
   // Var
   struct request_info* ri;
   unsigned long        tno = rp->hdr.TNO;
   unsigned short       size;

   // Validata
   if (!pSession->gw_logon_in_progress && rp->hdr.dst_id != pSession->our_node_id) {
      debug(("[0x%p]handleResponse: response NOT to us. Our id=%lu, packet id=%lu\n", pSession, pSession->our_node_id, rp->hdr.dst_id));
      freePacket(pSession, rp);
      return;
   }

   // Lock
   ENTER_CRIT;

   //try to find the outstanding request:
   ri = (struct request_info*)qPeekHead(&pSession->out_requests);
   while (ri) {
      if (ri->tno == tno) {
         //found!
         if (ri->data) {
            //data pointer supplied.
            size = DATALENGTH(rp);
            if (size) memcpy(ri->data, FRAMEDATA(rp), size);
            ri->data_len = size;
         } else {
            ri->data_len = 0;
         }
         ri->rc          = rp->hdr.cmd_rc;
         ri->gw_response = rp->hdr.gw_response ? 1 : 0;
         SEM_SIGNAL(ri->rxsignal);
         debug(("[0x%p]handleReponse TNO=%lu has been FOUND\n", pSession, rp->hdr.TNO));
         break;
      }
      ri=(struct request_info*)(ri->qlink.next);
   }

   // Unlock
   EXIT_CRIT;

   freePacket(pSession, rp);
}

/////////////////////////////////////////////////////////////////////////////
// Handles a received racp2 packet as a request.
//
// Input:
//      pSession   - Pointer to the session.
//      rp         - The racp2 packet received.
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
static void handleRequest(struct hSession* pSession, struct racp2_packet* rp)
{
   // Var
   struct client_info*  ci;
   struct racp2_packet* crp;
   struct racp2_packet* ptail;
   unsigned long        node_id, tno;
   unsigned short       ret, channel, size, datalen, headlen;
   int                  found;

   node_id = rp->hdr.src_id;
   tno     = rp->hdr.TNO;
   channel = rp->hdr.channel;
   debug(("[0x%p]handleRequest TNO=%lu, src_id=%lu, ch=%d\n", pSession, tno, node_id, channel));

   if (rp->hdr.dst_id != pSession->our_node_id) {
      debug(("[0x%p]handleRequest: request NOT to us. Our id=%lu, packet id=%lu\n", pSession, pSession->our_node_id, rp->hdr.dst_id));
      freePacket(pSession, rp);
      return;
   }

   //find client in client_list:
   found = 0;
   ENTER_CRIT;
   ci = (struct client_info*)qPeekHead(&pSession->client_list);
   while (ci) {
      if (node_id == ci->node_id) {
         //found!
         found = 1;
         break;
      }
      ci = (struct client_info*)(ci->qlink.next);
   }
   EXIT_CRIT;

   if (!found) {
      debug(("[0x%p]handleRequest: client for node_id=%lu was NOT found\n", pSession, node_id));

      //client node not found - we need to allocate:
      ENTER_CRIT;
      ci = (struct client_info*)qRemoveHead(&pSession->client_freelist);
      if (!ci) {
         //no client info structure available - we grab the LRU descriptor:
         ci = (struct client_info*)qSelRemove(&pSession->client_list, qPeekTail(&pSession->client_list));
         if (!ci) {
            //this is shit!
            EXIT_CRIT;
            freePacket(pSession, rp);
            goto bail_out;
         }
         //we have an old client info structure that needs to be cleaned up:
         while ((crp = (struct racp2_packet*)qRemoveHead(&ci->cacheq))) {
            freePacket(pSession, crp);
         }
      }
      EXIT_CRIT;

      //okay - we got a client_info for our use:
      debug(("[0x%p]handleRequest. Client %lu has been registered\n", pSession, node_id));
      qInit(&ci->cacheq);
      ci->last_used = 0;
      ci->node_id   = node_id;

      ENTER_CRIT;
      qInsertHead(&pSession->client_list, (qnode*)ci);
      EXIT_CRIT;
   }

   //okay, we got our descriptor with the info we need.
   //
   //look into our cache to see if this is an old request:
   found=0;
   ENTER_CRIT;
   ptail = (struct racp2_packet*)qPeekTail(&ci->cacheq); 
   crp   = (struct racp2_packet*)qPeekHead(&ci->cacheq);
   while (crp) {
      if (tno == crp->hdr.TNO) {
         debug(("[0x%p]handleRequest. Response for TNO=%lu found in cache.\n", pSession, tno));

         //found! - send back the cached response to the client:
         qSelRemove(&ci->cacheq, (qnode*)crp);
         EXIT_CRIT;
         crp->hdr.src_id = pSession->our_node_id; //this has been overwritten by the qlink.

         // Get size
         datalen = DATALENGTH(crp);
         headlen = HEADLENGTH(crp);

         // Is data encrypted?
         if (crp->hdr.encrypt) {
            size  = datalen / 16;
            if (datalen % 16 > 0) size++;
            size *= 16;
         } else {
            size  = datalen;
         }

         //let it go!
         ret = gwldSend(pSession->conn, (void*)&crp->hdr, headlen + size);
         if (ret) {
            //error!!!
            debug(("[0x%p]gwldSend failed. rc=%d. Closing connection\n", pSession, ret));
            gwldDisconnect(pSession->conn);
            pSession->gw_connection = 0;
         }

         ENTER_CRIT;
         qInsertHead(&ci->cacheq, (qnode*)crp);
         if (ptail != (struct racp2_packet*)qPeekTail(&ci->cacheq)) {
            //new tail. update timestamp:
            ci->last_used = pSession->time_tick;
         }
         found = 1;
         freePacket(pSession, rp);
         break;
      }
      crp = (struct racp2_packet*)((qnode*)(crp))->next;
   }
   EXIT_CRIT;

   if (!found) {
      //not found in cache:
      if (channel == 7 && rp->hdr.cmd_rc == 0x3fff) {
         debug(("[0x%p]handleRequest: Keep-alive request received. TNO=%lu\n", pSession, tno));
         gwmSendResponse(pSession, rp->hdr.src_id, tno, 0, "!", 1);
         freePacket(pSession, rp);
      } else {
         //this was a new request - send it up:
         debug(("[0x%p]handleRequest. New request TNO=%lu has been submitted to the application on ch=%d\n", pSession, tno, channel));
         //we save the src_id in the dst_id field because src_id will be overwritten by the q-routines.
         //this is a little dirty!
         rp->hdr.dst_id = rp->hdr.src_id;
         ENTER_CRIT;
         qInsertTail(&pSession->in_requests[channel], (qnode*)rp);
         EXIT_CRIT;
         SEM_SIGNAL(pSession->in_requests_notempty[channel]);
      }
   }

   bail_out:;
   //we are done!
}

/////////////////////////////////////////////////////////////////////////////
// Makes a logon transaction (request/response) with the gateway.
//
// Input:
//      pSession     - Pointer to the session.
//      nodeid       - The requested nodeid.
//      key          - The gateway logon key.
//
// Output:
//      rc           - The return code from the gateway.
//
// Returns:
//      GWRC_OK      - Success.
//      GWRC_ERROR   - Failed.
//
/////////////////////////////////////////////////////////////////////////////
static unsigned short Negotiation(struct hSession* pSession)
{
   // Var
   unsigned char        frame[16], opt_count;
   struct request_info  req;
   unsigned short       ret, size;
   long                 val;

   // Build frame
   opt_count         = 0;
   size              = 1;
   if (pSession->crypt_key_set) {
      frame[size++]  = 1;  // Option type: Encryption
      frame[size++]  = 1;  // Size: 1 Byte
      frame[size++]  = pSession->crypt_key_set ? 1 : 0;
      opt_count++;
   }
   if (pSession->comp_enable) {
      frame[size++]  = 2;  // Option type: Compression
      frame[size++]  = 1;  // Size: 1 Byte
      frame[size++]  = pSession->comp_enable ? 1 : 0;
      opt_count++;
   }
   if (pSession->ext_maxsize) {
      frame[size++]  = 3;  // Option type: Extended frames
      frame[size++]  = 4;  // Size: 4 Byte
      val            = pSession->ext_maxsize;
      memcpy(&frame[size], &val, 4);
      size          += 4;
      opt_count++;
   }
   frame[0]          = opt_count;

   // Fill out the outstanding request structure:
   ENTER_CRIT;
   req.tno  = pSession->next_tx_tno++;
   EXIT_CRIT;
   req.data = frame;
   SEM_CREATE(req.rxsignal, 0);
   ENTER_CRIT;
   qInsertTail(&pSession->out_requests, (qnode*)&req);
   EXIT_CRIT;

   // Fill out the packet to send:
   pSession->req_packet_std.hdr.src_id        =   0;
   pSession->req_packet_std.hdr.dst_id        =   0;
   pSession->req_packet_std.hdr.crc           =   0;
   pSession->req_packet_std.hdr.TNO           =   req.tno;
   pSession->req_packet_std.hdr.data_len      =   size;
   pSession->req_packet_std.hdr.channel       =   0;
   pSession->req_packet_std.hdr.encrypt       =   0;
   pSession->req_packet_std.hdr.compress      =   0;
   pSession->req_packet_std.hdr.gw_response   =   0;
   pSession->req_packet_std.hdr.cmd_rc        =   0x000A;
   memcpy(pSession->req_packet_std.data, frame, size);
   pSession->req_packet_std.hdr.crc           =   gw_calc_crc16data(0, (unsigned char*)&pSession->req_packet_std.hdr, sizeof(struct racp2_header) + pSession->req_packet_std.hdr.data_len);
   debug(("[0x%p]Negotiation: TNO=%lu, crc=0x%x!\n", pSession, req.tno, pSession->req_packet_std.hdr.crc));

   // Send packet
   ret = gwldSend(pSession->conn, (void*)&pSession->req_packet_std.hdr, sizeof(struct racp2_header) + size);
   if (ret) {
      ENTER_CRIT;
      qSelRemove(&pSession->out_requests, (qnode*)&req);
      EXIT_CRIT;
      SEM_CLOSE(req.rxsignal);
      return GWRC_ERROR;
   }

   // Wait for answer to arrive:
   if (SEM_WAIT_T(req.rxsignal, RESPONSE_TIMEOUT)) {
      debug(("[0x%p]Negotiation: Timeout waiting for response. TNO=%lu!\n", pSession, req.tno));

      //timeout:
      ENTER_CRIT;
      qSelRemove(&pSession->out_requests, (qnode*)&req);
      EXIT_CRIT;
      SEM_CLOSE(req.rxsignal);
      return GWRC_ERROR;
   } else {
      debug(("[0x%p]Negotiation: Response has arrived. TNO=%lu!\n", pSession, req.tno));

      //response has arrived:
      ENTER_CRIT;
      qSelRemove(&pSession->out_requests, (qnode*)&req);
      EXIT_CRIT;
      SEM_CLOSE(req.rxsignal);

      // Parse options
      opt_count         = frame[0];
      size              = 1;
      for (;opt_count;opt_count--)
      {
         // Handle option
         switch (frame[size])
         {
            case 1:  // Encryption
               if (frame[size + 2]) pSession->gw_apply_encrypt = 1;
               debug(("[0x%p]Negotiation: Encryption=%d\n", pSession, pSession->gw_apply_encrypt));
               size += frame[size + 1] + 2;
               break;

            case 2:  // Compression
               if (frame[size + 2]) pSession->gw_apply_compress = 1;
               size += frame[size + 1] + 2;
               debug(("[0x%p]Negotiation: Compression=%d\n", pSession, pSession->gw_apply_compress));
               break;

            case 3:  // Extended frames
               // Get extended frame size
               memcpy(&val, &frame[size + 2], frame[size + 1]);

               // Enable
               if (val) pSession->gw_extended_maxsize = (unsigned short)(val & 0x0000FFFF);
               debug(("[0x%p]Negotiation: Extended frames, max=%d\n", pSession, pSession->gw_extended_maxsize));

               // Next
               size += frame[size + 1] + 2;
               break;

            default: // Unknown option
               debug(("[0x%p]Negotiation: Unknown option (type=%d)\n", pSession, frame[size]));
               size += frame[size + 1] + 2;
               break;
         }
      }

      // Completed
      return GWRC_OK;
   }
}

/////////////////////////////////////////////////////////////////////////////
// Makes a logon transaction (request/response) with the gateway.
//
// Input:
//      pSession     - Pointer to the session.
//      nodeid       - The requested nodeid.
//      key          - The gateway logon key.
//
// Output:
//      rc           - The return code from the gateway.
//
// Returns:
//      GWRC_OK      - Success.
//      GWRC_ERROR   - Failed.
//
/////////////////////////////////////////////////////////////////////////////
static unsigned short logonTransaction(
                                       struct hSession* pSession,
                                       unsigned long    nodeid,
                                       const char*      key,
                                       unsigned short*  rc
                                      )
{
   // Var
   unsigned char        logon[16];
   struct request_info  req;
   unsigned short       ret, size;

   // Build logon frame
   memcpy(&logon[0], &nodeid, 4);                      // Requested Node ID
   strcpy((char*)&logon[4], key);                      // Logon key
   size      = 13;                                     // Size of frame

   // Fill out the outstanding request structure:
   ENTER_CRIT;
   req.tno  = pSession->next_tx_tno++;
   EXIT_CRIT;
   req.data = logon;
   SEM_CREATE(req.rxsignal, 0);
   ENTER_CRIT;
   qInsertTail(&pSession->out_requests, (qnode*)&req);
   EXIT_CRIT;

   // Fill out the packet to send:
   pSession->req_packet_std.hdr.src_id        =   0;
   pSession->req_packet_std.hdr.dst_id        =   0;
   pSession->req_packet_std.hdr.crc           =   0;
   pSession->req_packet_std.hdr.TNO           =   req.tno;
   pSession->req_packet_std.hdr.data_len      =   size;
   pSession->req_packet_std.hdr.channel       =   0;
   pSession->req_packet_std.hdr.encrypt       =   0;
   pSession->req_packet_std.hdr.compress      =   0;
   pSession->req_packet_std.hdr.gw_response   =   0;
   pSession->req_packet_std.hdr.cmd_rc        =   1;
   memcpy(pSession->req_packet_std.data, logon, size);

   pSession->req_packet_std.hdr.crc           =   gw_calc_crc16data(0, (unsigned char*)&pSession->req_packet_std.hdr, sizeof(struct racp2_header) + pSession->req_packet_std.hdr.data_len);
   debug(("[0x%p]logonTransaction: TNO=%lu, crc=0x%x!\n", pSession, req.tno, pSession->req_packet_std.hdr.crc));

   // Encrypt packet
   if (pSession->gw_apply_encrypt) {
      rijndaelEncryptBuffer128(pSession->crypt_key, logon, pSession->req_packet_std.data, 1);
      pSession->req_packet_std.hdr.encrypt = 1;
      size                                 = 16;
      debug(("[0x%p]logonTransaction: Encrypt packet\n", pSession));
   }

   // Send packet
   ret = gwldSend(pSession->conn, (void*)&pSession->req_packet_std.hdr, sizeof(struct racp2_header) + size);
   if (ret) {
      ENTER_CRIT;
      qSelRemove(&pSession->out_requests, (qnode*)&req);
      EXIT_CRIT;
      SEM_CLOSE(req.rxsignal);
      return GWRC_ERROR;
   }

   pSession->logon_rxsignal = &req.rxsignal;

   // Wait for answer to arrive:
   if (SEM_WAIT_T(req.rxsignal, RESPONSE_TIMEOUT)) {
      debug(("[0x%p]logonTransaction: Timeout waiting for response. TNO=%lu!\n", pSession, req.tno));
      pSession->logon_rxsignal = 0;

      //timeout:
      ENTER_CRIT;
      qSelRemove(&pSession->out_requests, (qnode*)&req);
      EXIT_CRIT;
      SEM_CLOSE(req.rxsignal);
      return GWRC_ERROR;
   } else {
      pSession->logon_rxsignal = 0;
      if (pSession->logon_cancelled) {
         debug(("[0x%p]logonTransaction: Logon cancelled\n", pSession));
         ENTER_CRIT;
         qSelRemove(&pSession->out_requests, (qnode*)&req);
         EXIT_CRIT;
         SEM_CLOSE(req.rxsignal);
         return GWRC_ERROR;
      }
      debug(("[0x%p]logonTransaction: Response has arrived. TNO=%lu!\n", pSession, req.tno));

      //response has arrived:
      ENTER_CRIT;
      qSelRemove(&pSession->out_requests, (qnode*)&req);
      EXIT_CRIT;
      SEM_CLOSE(req.rxsignal);

      *rc = req.rc & ~0x8000;
      if (!*rc) {
         //success.
         // Set NodeId
         if (pSession->initial_node_id == 0)
            pSession->our_node_id = *(unsigned long*)logon;
         else
            pSession->our_node_id = pSession->initial_node_id;
      }
      return GWRC_OK;
   }
}

/////////////////////////////////////////////////////////////////////////////
// Makes a logon transaction (request/response) with the gateway.
// Opens for Compression and Encryption of data send and received from
// the Gateway.
//
// Input:
//      pSession     - Pointer to the session.
//
// Output:
//      rp           - The return code from the gateway.
//
// Returns:
//      GWRC_OK      - Success.
//      GWRC_ERROR   - Failed.
//
/////////////////////////////////////////////////////////////////////////////
static unsigned short logonTransactionAdv(struct hSession* pSession,
                                          unsigned short*  rc)
{
   // Var
   unsigned char        logon[16];
   struct request_info  req;
   unsigned short       ret, size;

   // Build logon frame
   memcpy(&logon[0], &pSession->initial_node_id, 4);   // Requested Node ID
   strcpy((char*)&logon[4], pSession->gw_key);         // Logon key
   logon[13] = pSession->crypt_key_set ? 1 : 0;        // Encryption level supported by client
   logon[14] = pSession->comp_enable ? 1 : 0;          // Compression level supported by client
   size      = 15;                                     // Size of frame

   // Fill out the outstanding request structure:
   ENTER_CRIT;
   req.tno  = pSession->next_tx_tno++;
   EXIT_CRIT;
   req.data = logon;
   SEM_CREATE(req.rxsignal, 0);
   ENTER_CRIT;
   qInsertTail(&pSession->out_requests, (qnode*)&req);
   EXIT_CRIT;

   // Fill out the packet to send:
   pSession->req_packet_std.hdr.src_id        =   0;
   pSession->req_packet_std.hdr.dst_id        =   0;
   pSession->req_packet_std.hdr.crc           =   0;
   pSession->req_packet_std.hdr.TNO           =   req.tno;
   pSession->req_packet_std.hdr.data_len      =   size;
   pSession->req_packet_std.hdr.channel       =   0;
   pSession->req_packet_std.hdr.encrypt       =   0;
   pSession->req_packet_std.hdr.compress      =   0;
   pSession->req_packet_std.hdr.gw_response   =   0;
   pSession->req_packet_std.hdr.cmd_rc        =   2;
   memcpy(pSession->req_packet_std.data, logon, size);
   pSession->req_packet_std.hdr.crc           =   gw_calc_crc16data(0, (unsigned char*)&pSession->req_packet_std.hdr, sizeof(struct racp2_header) + pSession->req_packet_std.hdr.data_len);
   debug(("[0x%p]logonTransactionAdv: TNO=%lu, crc=0x%x!\n", pSession, req.tno, pSession->req_packet_std.hdr.crc));

   // Encrypt packet
   if (pSession->crypt_key_set) {
      rijndaelEncryptBuffer128(pSession->crypt_key, logon, pSession->req_packet_std.data, 1);
      pSession->req_packet_std.hdr.encrypt = 1;
      size                                 = 16;
      debug(("[0x%p]logonTransactionAdv: Encrypt packet\n", pSession));
   }

   // Send packet
   ret = gwldSend(pSession->conn, (void*)&pSession->req_packet_std.hdr, sizeof(struct racp2_header) + size);
   if (ret) {
      ENTER_CRIT;
      qSelRemove(&pSession->out_requests, (qnode*)&req);
      EXIT_CRIT;
      SEM_CLOSE(req.rxsignal);
      return GWRC_ERROR;
   }

   pSession->logon_rxsignal = &req.rxsignal;

   // Wait for answer to arrive:
   if (SEM_WAIT_T(req.rxsignal, RESPONSE_TIMEOUT)) {
      debug(("[0x%p]logonTransactionAdv: Timeout waiting for response. TNO=%lu!\n", pSession, req.tno));
      pSession->logon_rxsignal = 0;

      //timeout:
      ENTER_CRIT;
      qSelRemove(&pSession->out_requests, (qnode*)&req);
      EXIT_CRIT;
      SEM_CLOSE(req.rxsignal);
      return GWRC_ERROR;
   } else {
      pSession->logon_rxsignal = 0;
      if (pSession->logon_cancelled) {
         debug(("[0x%p]logonTransactionAdv: Logon cancelled\n", pSession));
         ENTER_CRIT;
         qSelRemove(&pSession->out_requests, (qnode*)&req);
         EXIT_CRIT;
         SEM_CLOSE(req.rxsignal);
         return GWRC_ERROR;
      }
      debug(("[0x%p]logonTransactionAdv: Response has arrived. TNO=%lu!\n", pSession, req.tno));

      //response has arrived:
      ENTER_CRIT;
      qSelRemove(&pSession->out_requests, (qnode*)&req);
      EXIT_CRIT;
      SEM_CLOSE(req.rxsignal);

      *rc = req.rc & ~0x8000;
      if (!*rc) {
         //success.
         // Set NodeId
         if (pSession->initial_node_id == 0)
            pSession->our_node_id = *(unsigned long*)logon;
         else
            pSession->our_node_id = pSession->initial_node_id;
         // Advanced data
         pSession->gw_apply_compress = logon[5];
         pSession->gw_apply_encrypt  = logon[4];
         debug(("[0x%p]logonTransactionAdv: Encryption=%d, Compression=%d\n", pSession, logon[4], logon[5]));
      }
      return GWRC_OK;
   }
}

/////////////////////////////////////////////////////////////////////////////
// This thread monitors the connection status.
//
// Input:
//      pSession     - Pointer to the session.
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
static void thConnectionMonitor(struct hSession *pSession)
{
   // Var
   unsigned short state, rc, trc, activity, con_retry, con_timer, throttle;
   unsigned char  inet, legacy;

   pSession->running_threads++;
   debug(("[0x%p]thConnectionMonitor running.\n", pSession));

   state     = 0;
   activity  = 0;
   con_retry = 0;
   con_timer = 0; 

   // Thread main loop
   do {
      SLEEP(250);

      ageResponseCache(pSession);

      //connection monitoring:
      debug(("[0x%p]thConnectionMonitor. state=%d, activity=%d\n", pSession, state, activity));

      if (!pSession->gw_connection && (state == 2 || state == 3 || state == 4)) {
         //state 2,3,4 is the states where the connection can be closed from outside this state-machine.
         debug(("[0x%p]thConnectionMonitor: Connection closed from outside SM. going to state 1\n", pSession));
         activity = 0;
         state    = 1;
      }
      inet = gwldIsInternetPresent(pSession->conn) == GWRC_OK;

      if (!inet && (state == 2 || state == 3 || state == 4 || state == 5)) {
         debug(("[0x%p]thConnectionMonitor: Internet connection dropped outside SM. going to state 1\n", pSession));
         gwldDisconnect(pSession->conn);
         pSession->gw_connection = 0;
         state = 1; 
      }

      if (state != 0 && !pSession->opened) {
         //gwmClose() has been called!
         if (pSession->gw_connection) {
            gwldDisconnect(pSession->conn);
            pSession->gw_connection = 0;
         }
         state = 0;
         debug(("[0x%p]thConnectionMonitor: Connection has been closed outside SM. going to state 0\n", pSession));
      }

      switch (state) {
      case 0: //waiting for open.
         if (pSession->opened) {
            activity = 0;
            throttle = 0;
            state    = 1;
         }
         break;

      case 1: //waiting for internet available and initiating connection to gateway.
         if (!inet || !gwldIsInternetSuspended(pSession->conn)) {
            activity = 0;
            throttle = 0;
            break;
         }

         // Reset connection options
         pSession->gw_apply_compress   = 0;
         pSession->gw_apply_encrypt    = 0;
         pSession->gw_extended_maxsize = 0;

         // Throttle connection attempts
         if (throttle && --throttle) {
            break;
         }
         throttle = 20; // Wait 5 seconds

         if (activity++ >= MAX_CONNECTION_RETRIES) {
#if THROTTLE_ENABLE==1
            if (++con_retry > (GW_CON_MAX_DELAY / GW_CON_DELAY_STEP)) con_retry = 1;
            con_timer = con_retry * (GW_CON_DELAY_STEP * 1000 / TIMER_TICK_FRQ);
            debug(("thConnectionMonitor - Max. retry. Timed close=%d\n", con_timer));
            activity = 0;
            state    = 5;
#else
            //we can not reconnect - we need to drop the internet connection.
            debug(("[0x%p]thConnectionMonitor - Disconnect internet\n", pSession));
            gwldDisconnectInternet(pSession->conn);
            activity = 0;
            state    = 0;
#endif
            break;
         }

         //not connected to gateway!
         debug(("[0x%p]thConnectionMonitor - establishing connection.\n", pSession));
         rc = gwldConnect(pSession->conn, pSession->gw_address, pSession->gw_port, pSession->tls_server, pSession->tls_cert, pSession->tls_key, pSession->tls_pass);
         if (rc) {
            debug(("[0x%p]thConnectionMonitor - gwldConnect failed rc=%d\n", pSession, rc));
            break;
         }
         debug(("[0x%p]thConnectionMonitor - attempting gateway LOGON.\n", pSession));

         pSession->gw_logon_in_progress = 1;
         pSession->gw_connection        = 1;
         legacy                         = 1;

         if (pSession->ext_maxsize) {
            debug(("[0x%p]thConnectionMonitor - extended frames is enabled.\n", pSession));

            // Request gateway version
            rc = logonTransaction(
                                  pSession,
                                  0,
                                  GW_VER_REQUEST,
                                  &trc
                                 );
            if (rc) {
               debug(("[0x%p]thConnectionMonitor - gateway version failed.\n", pSession));
               pSession->gw_connection        = 0;
               pSession->gw_logon_in_progress = 0;
               gwldDisconnect(pSession->conn);
               break;
            }
            if (trc > 8000) trc -= 8000;
            if (trc<320) {
               debug(("[0x%p]thConnectionMonitor - gateway do not support extended frames.\n", pSession));
            } else {
               // Negotiate options
               rc = Negotiation(pSession);
               if (rc) {
                  debug(("[0x%p]thConnectionMonitor - configuration negotiation failed.\n", pSession));
                  pSession->gw_connection        = 0;
                  pSession->gw_logon_in_progress = 0;
                  gwldDisconnect(pSession->conn);
                  break;
               }

               // Logon
               legacy = 0;
               rc     = logonTransaction(
                                         pSession,
                                         pSession->initial_node_id,
                                         pSession->gw_key,
                                         &trc
                                        );
            }
         }

         if (legacy) {
            rc = logonTransactionAdv(pSession, &trc);
         }
         if (rc) {
            debug(("[0x%p]thConnectionMonitor - gateway LOGON failed.\n", pSession));
            pSession->gw_connection        = 0;
            pSession->gw_logon_in_progress = 0;
            gwldDisconnect(pSession->conn);
            break;
         }
         if (trc) {
            debug(("[0x%p]thConnectionMonitor - gateway LOGON failed. logon rc=%d\n", pSession, trc));
            pSession->gw_connection        = 0;
            pSession->gw_logon_in_progress = 0;

            //this is a special case where we know we have communicated with the gateway, but
            //most likely because of wrong KEY the logon has failed.
            //we floors the activity counter because we do not want the internet connection to be
            //dropped in this case.
            activity = 0;
            gwldDisconnect(pSession->conn);
            break;
         }
         debug(("[0x%p]thConnectionMonitor - gateway LOGON SUCCESSFUL. Our nodeID=%d\n", pSession, pSession->our_node_id));
         pSession->gw_logon_in_progress = 0;
         activity  = 0;
         con_retry = 0;
         throttle  = 0;
         state     = 2;
         break;

      case 2: //connection established - monitor activity:
         if (ACTIVITY_MONITOR_THRESHOLD) {  //can be disabled..
            if (pSession->userTransactionsOK) {
               debug(("[0x%p]thConnectionMonitor - user transaction successful.\n", pSession));
               activity = 0;
               pSession->userTransactionsOK = 0;
            }
            if (gwldIsInternetSuspended(pSession->conn)) {
               //only count if internet is not suspended..
               if (++throttle == 20) {
                  throttle = 0;
                  if (++activity == ACTIVITY_MONITOR_THRESHOLD) {
                     activity = 0;
                     state    = 3;
                  }
               }
            }
         }
         if (pSession->forceSelfTransaction) {
            debug(("[0x%p]thConnectionMonitor - self-transaction has been forced\n", pSession)); 
            pSession->forceSelfTransaction = 0;
            state = 3;
         }
         break;

      case 3: //first attempt to make a transaction.
         debug(("[0x%p]thConnectionMonitor - Performing self transaction #1\n", pSession));
         rc = gwmTransaction(pSession, 7, pSession->our_node_id, 0x3fff, "?", 1, &trc, &trc, &trc);
         if (rc == GWRC_OK) {
            activity = 0; state = 2;
         } else {
            activity = 0; state = 4;
         }
         debug(("[0x%p]thConnectionMonitor - Self transaction #1 rc=%d\n", pSession, rc));
         break;

      case 4: //second attempt to make a transaction.
         gwldDisconnect(pSession->conn);
         pSession->gw_connection = 0;
         activity = 0;
         state    = 1;
         break;

#if THROTTLE_ENABLE==1
      case 5:  //throttle before disconnect of internet.
         debug(("thConnectionMonitor - Throttle=%d\n",con_timer));
         if (--con_timer == 0) {
            //we can not reconnect - we need to drop the internet connection.
            debug(("thConnectionMonitor - Disconnect internet\n"));
            gwldDisconnectInternet(pSession->conn);
            activity = 0;
            state    = 0;
         }
         break;
#endif
      }
   }
   while (pSession->initialized);

   debug(("[0x%p]thConnectionMonitor - stopped\n", pSession));
   pSession->running_threads--;
}

/////////////////////////////////////////////////////////////////////////////
// This thread handles incoming data.
//
// Input:
//      pSession     - Pointer to the session.
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
static void thReceiver(struct hSession *pSession)
{
   // Var
   struct racp2_packet* rp;
   unsigned short       rc, size, blocks=0;

   pSession->running_threads++;
   debug(("[0x%p]thReceiver running.\n", pSession));

   // Thread main loop
   while (pSession->initialized) {
      // Wait for connection is open
      while (pSession->initialized && !pSession->gw_connection) SLEEP(100);
      if (!pSession->initialized) break;

      rp = allocPacket(pSession, 1);
      if (!rp) {
         debug(("[0x%p]thReceiver: failed allocating racp2 packet.\n", pSession));
         SLEEP(1000);
         continue;
      }

      //debug(("thReceiver: now receiving header.\n"));
      rc = blockReceive(pSession, &rp->hdr, sizeof(rp->hdr));
      if (!rc) {
         if (CHECKEXT(rp)) {
            // Var
            struct ext_packet* temp;
            unsigned char*     data;
            unsigned short     size;

            // Extended frame received
            temp = (struct ext_packet*)allocPacket(pSession, 2);
            if (temp==NULL)
            {
               // No extended frames, What to do????
               debug(("[0x%p]thReceiver: No extended frames free. Closing connection\n", pSession));
               freePacket(pSession, rp);
               gwldDisconnect(pSession->conn);
               pSession->gw_connection = 0;
               continue;
            }
            debug(("[0x%p]thReceiver: Switch to extended frame\n", pSession));

            // Swap header information
            memcpy(&temp->hdr, &rp->hdr, sizeof(rp->hdr));

            // Calculate remainder
            data = (unsigned char*)&temp->hdr;
            data = &data[sizeof(rp->hdr)];
            size = HEADLENGTH(temp) - sizeof(rp->hdr);

            // Switch buffer
            freePacket(pSession, rp);
            rp = (struct racp2_packet*)temp;

            // Read the remaining header
            rc = blockReceive(pSession, data, size);
         }
      }
      if (!rc) {
         // Is packet encrypted?
         if (rp->hdr.encrypt) {
            // Calculate data length
            size   = DATALENGTH(rp);
            blocks = size / 16;
            if (size % 16 > 0) blocks++; // Round fractions up
            size   = blocks * 16;
         } else {
            size   = DATALENGTH(rp);
         }
         debug(("[0x%p]thReceiver: now receiving data len=%d\n", pSession,size));
         rc = size <= MAXLENGTH(rp) ? blockReceive(pSession, FRAMEDATA(rp), size) : 1;
      }
      if (rc) {
         if (pSession->gw_logon_in_progress) {
            debug(("[0x%p]thReceiver: failure receiving data during logon - resuming operation\n", pSession));
            rp->hdr.channel     = 0;
            rp->hdr.cmd_rc      = 0x8003;
            rp->hdr.compress    = 0;
            rp->hdr.crc         = 0;
            rp->hdr.data_len    = 0;
            rp->hdr.dst_id      = 0;
            rp->hdr.encrypt     = 0;
            rp->hdr.gw_response = 1;
            rp->hdr.src_id      = 0;
            rp->hdr.TNO         = pSession->next_tx_tno - 1;
         } else {
            //failed!
            debug(("[0x%p]thReceiver: failure receiving data. Closing connection\n", pSession));
            freePacket(pSession, rp);
            gwldDisconnect(pSession->conn);
            pSession->gw_connection = 0;
            continue;
         }
      }

      // Is packet encrypted?
      if (rp->hdr.encrypt) {
         // Thread safety
         SEM_WAIT(pSession->comp_mutex);
            // Decrypt data
            rijndaelDecryptBuffer128(pSession->crypt_key, FRAMEDATA(rp), (unsigned char*)pSession->comp_buf, blocks);
            // Use decrypted data
            memcpy(FRAMEDATA(rp), pSession->comp_buf, DATALENGTH(rp));
            rp->hdr.encrypt = 0;
         // Thread safety
         SEM_SIGNAL(pSession->comp_mutex);
      }

      // Is packet compressed?
      if (rp->hdr.compress) {
         // Thread safety
         SEM_WAIT(pSession->comp_mutex);
            // Expand data
            size             = expand_block(FRAMEDATA(rp), (unsigned char*)pSession->comp_buf, DATALENGTH(rp));
            // Use Expanded data
            memcpy(FRAMEDATA(rp), pSession->comp_buf, size);
            rp->hdr.compress = 0;
            SET_LENGTH(rp, size);
         // Thread safety
         SEM_SIGNAL(pSession->comp_mutex);
      }

      //we now have a packet:
      //check CRC:
      if (checkCRC(rp)) {
         //error!
         debug(("[0x%p]thReceiver: crc error. Closing connection\n", pSession));
         freePacket(pSession, rp);
         gwldDisconnect(pSession->conn);
         pSession->gw_connection = 0;
         continue;
      }

      if (rp->hdr.channel > 7) {
         debug(("[0x%p]thReceiver: invalid channel.  ch=%d. Closing connection\n", pSession, rp->hdr.channel));
         freePacket(pSession, rp);
         gwldDisconnect(pSession->conn);
         pSession->gw_connection = 0;
         continue;
      }
      if (rp->hdr.cmd_rc & 0x8000) {
         //response packet has arrived:
         handleResponse(pSession, rp);
      } else {
         //request packet has arrived:
         handleRequest(pSession, rp);
      }
   }
   debug(("[0x%p]thReceiver stopped.\n", pSession));
   pSession->running_threads--;
}

/////////////////////////////////////////////////////////////////////////////
// This thread handles the timer.
//
// Input:
//      pSession     - Pointer to the session.
//
// Output:
//      None.
//
// Returns:
//      None.
//
/////////////////////////////////////////////////////////////////////////////
static void thTimertick(struct hSession *pSession)
{
   long timeout = 0;

   pSession->running_threads++;
   debug(("[0x%p]thTimertick running\n", pSession));

   // Thread main loop
   for (;pSession->initialized;) {
      SLEEP(100);
      timeout += 100;
      if (timeout >= TIMER_TICK_FRQ) {
         timeout = 0;
         pSession->time_tick++;
      }
   }

   debug(("[0x%p]thTimertick stopped\n", pSession));
   pSession->running_threads--;
}


/*****************************************************************************
**
** EXPORTED FUNCTIONS 
**
*****************************************************************************/

void gwInitialize(void)
{
   if (initialized) return;

   INIT_CRIT;
   INIT_RANDOM;

   initialized=1;
   debug(("gwInitialize - first time initialization\n"));
   debug(("gwInitialize - std packets=%d\n", GWPROT_MAX_PACKETS_STD));
   debug(("gwInitialize - ext packets=%d\n", GWPROT_MAX_PACKETS_EXT));
   debug(("gwInitialize - max client =%d\n", GWPROT_MAX_CLIENTS));
}

/////////////////////////////////////////////////////////////////////////////
// Query whether the gateway communication module has been initialized
//
// Input:
//      None.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK       - Gateway communication module has been initialized
//     GWRC_NOT_INIT - Gateway communication module has NOT been initialized 
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwIsInitialized(void)
{
   return initialized ? GWRC_OK : GWRC_NOT_INIT;
}

/////////////////////////////////////////////////////////////////////////////
// Retrieve the version number of the GWPROT library.
//
// Input:
//      None.
//
// Output:
//      None.
//
// Returns:
//     The version scaled by 100.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwGetVer(void)
{
   return 620;
}

/////////////////////////////////////////////////////////////////////////////
// Create a new gateway communication session, this function will allocate
// and initialize a new session which can be used with the gwm* function.
//
// When complete with the session it must be freed with gwmDelete.
//
// Output:
//      pSession        - the session handle.
//
// Returns:
//      GWRC_OK          - Success.
//      GWRC_ERROR       - General error.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmCreate(struct hSession **pSession)
{
   // Var
   int j;

   // Allocate session structure
   *pSession = (struct hSession *)malloc(sizeof(struct hSession));
   if (*pSession == NULL) return GWRC_ERROR;

   // Initialize handle
   memset(*pSession, 0, sizeof(struct hSession));
   gwldCreate(&(*pSession)->conn);
   debug(("[0x%p]gwmCreate - new session created\n", *pSession));

   // Set default configuration
   (*pSession)->max_connection_attempt = 3;
   (*pSession)->max_send_req_attempt   = 5;
   (*pSession)->response_timeout       = 45000;
   (*pSession)->alive_freq             = 300 / (TIMER_TICK_FRQ / 1000);
   (*pSession)->comp_enable            = 1;   // Compression is enabled as default
   (*pSession)->ext_maxsize            = 0;

   // Initialize session
   SEM_CREATE((*pSession)->send_mutex, 1);
   SEM_CREATE((*pSession)->comp_mutex, 1);

   qInit(&(*pSession)->out_requests);
   for (j = 0; j < GW_MAXCHANNELS; j++) {
      qInit(&(*pSession)->in_requests[j]);
      SEM_CREATE((*pSession)->in_requests_notempty[j], 0);
   }

   //build free-list:
   qInit(&(*pSession)->std_freelist);
   for (j = 0; j < GWPROT_MAX_PACKETS_STD; j++) {
      (*pSession)->std_buffers[j].type = 1;
      qInsertHead(&(*pSession)->std_freelist, (qnode*)&(*pSession)->std_buffers[j].node);
   }
   SEM_CREATE((*pSession)->std_freelist_notempty, GWPROT_MAX_PACKETS_STD);
   qInit(&(*pSession)->ext_freelist);
   for (j = 0; j < GWPROT_MAX_PACKETS_EXT; j++) {
      (*pSession)->ext_buffers[j].type = 2;
      qInsertHead(&(*pSession)->ext_freelist, (qnode*)&(*pSession)->ext_buffers[j].node);
   }
   SEM_CREATE((*pSession)->ext_freelist_notempty, GWPROT_MAX_PACKETS_EXT);

   //initialize the client-list;
   qInit(&(*pSession)->client_list);
   qInit(&(*pSession)->client_freelist);
   for (j = 0; j < GWPROT_MAX_CLIENTS; j++) {
      qInsertHead(&(*pSession)->client_freelist, (qnode*)&(*pSession)->client_buffers[j]);
   }
   (*pSession)->next_tx_tno = GET_RANDOM();
   debug(("[0x%p]Initial next_tx_tno=%lu\n", (*pSession), (*pSession)->next_tx_tno));

   // Initialize request buffer
   (*pSession)->req_packet_ext.type            = 2;
   (*pSession)->req_packet_ext.hdr.encrypt     = 0;
   (*pSession)->req_packet_ext.hdr.compress    = 0;
   (*pSession)->req_packet_ext.hdr.gw_response = 0;
   (*pSession)->req_packet_ext.hdr.priority    = 0;
   (*pSession)->req_packet_ext.hdr.unack       = 0;
   (*pSession)->req_packet_ext.hdr.reserved1   = 0;
   (*pSession)->req_packet_ext.hdr.extpack     = 3;
   (*pSession)->req_packet_ext.hdr.reserved2   = 0;
   (*pSession)->req_packet_std.type            = 1;
   (*pSession)->req_packet_std.hdr.encrypt     = 0;
   (*pSession)->req_packet_std.hdr.compress    = 0;
   (*pSession)->req_packet_std.hdr.gw_response = 0;

   // Initialize TLS
   (*pSession)->tls_cert[0] = 0;

   (*pSession)->opened      = 0;
   (*pSession)->initialized = 1;

   // Start threads
   CREATETHREAD(thTimertick, (*pSession));
   CREATETHREAD(thConnectionMonitor, (*pSession));
   CREATETHREAD(thReceiver, (*pSession));

   // Completed
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// This will open a connection to a Gateway.
// Must be called before any other function in this DLL (except
// gwmSetParameters and gwmSetEncryptionKey).
//
// Input:
//      pSession     - Pointer to the session.
//      node_id      - Our NodeId. Interval: 0..999999. (0 = Dynamic NodeId).
//      ip_addr      - String containing Gateway IP address. (can be symbolic)
//      port         - The Gateway IP port.
//      key          - String containing Gateway key. (Max. 8 characters)
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK       - Success.
//     GWRC_IS_OPEN  - Gateway communication is already open.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmOpen(
                                struct hSession *pSession,
                                unsigned long   node_id,
                                const char      *ip_addr,
                                unsigned short  port,
                                const char      key[9]
                              )
{
   // Validate
   if (!pSession) return GWRC_ERROR;
   if (pSession->opened) return GWRC_IS_OPEN;

   // Clear connection
   pSession->gw_connection             = 0;
   pSession->gw_logon_in_progress      = 0;
   pSession->userTransactionsOK        = 0;
   pSession->forceSelfTransaction      = 0;
   pSession->logon_cancelled           = 0;

   // Set connection parameters
   pSession->initial_node_id           = node_id;
   pSession->gw_port                   = port;
   strncpy(pSession->gw_address, ip_addr, 42);
   strncpy(pSession->gw_key, key, 10);
   pSession->gw_address[41]            = 0;
   pSession->gw_key[9]                 = 0;

   // Completed
   pSession->opened                    = 1;
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Query whether the gateway communication is open.
//
// Input:
//      pSession     - Pointer to the session.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK       - Gateway communication is open.
//     GWRC_NOT_OPEN - Gateway communication is not open.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmIsOpened(struct hSession *pSession)
{
   if (!pSession) return GWRC_NOT_OPEN;
   return pSession->opened ? GWRC_OK : GWRC_NOT_OPEN;
}

/////////////////////////////////////////////////////////////////////////////
// Sets the TLS parameters of a session.
//
// Input:
//      pSession                   - Pointer to the session.
//      server                     - The path to the PEM file used to verify
//                                   the server certificate.
//      cert                       - The client certificate to present to
//                                   the server.
//      key                        - The private key of the client.
//      password                   - The password of the private key.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK        - Success.
//     GWRC_ERROR     - Invalid input parameter
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmSetTLSParameters(
                                            struct hSession* pSession,
                                            char*            server,
                                            char*            cert,
                                            char*            key,
                                            char*            password
                                           )
{
   // Validate
   if (!pSession) return GWRC_ERROR;
   if (cert && key == NULL) return GWRC_ERROR;
   if (key && cert == NULL) return GWRC_ERROR;
   if (server && !ValidateCertificate(server)) return GWRC_ERROR;
   if (cert && !ValidateClient(cert, key, password)) return GWRC_ERROR;

   // Set default
   pSession->tls_server[0] = 0;
   pSession->tls_cert[0]   = 0;
   pSession->tls_key[0]    = 0;
   pSession->tls_pass[0]   = 0;

   // Set config
   if (server && strlen(server) > 0) {
      strncpy(pSession->tls_server, server, 512);
      pSession->tls_server[511] = 0;
   }
   if (cert && strlen(cert) > 0) {
      strncpy(pSession->tls_cert, cert, 512);
      pSession->tls_cert[511] = 0;
   }
   if (key && strlen(key) > 0) {
      strncpy(pSession->tls_key, key, 512);
      pSession->tls_key[511] = 0;
   }
   if (password && strlen(password) > 0) {
      strncpy(pSession->tls_pass, password, 51);
      pSession->tls_pass[50] = 0;
   }

   // Completed
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Sets the Advanced connection parameters of a session.
//
// Input:
//      pSession                   - Pointer to the session.
//      gw_max_connection_attempt  - max number of connection attempts before GPRS re-connects.
//                                   interval: 1..60. (default: 3) (NOT USED ON PC IMPLEMENTATION)
//      gw_max_send_req_attempt    - max number of send-request attempt before send fails.
//                                   interval: 1..60 (default: 3).
//      gw_response_timeout        - time waiting for response in seconds.
//                                   interval: 5..60 (def: 30 seconds).
//      gw_alive_freq              - frequency for sending self-transactions in seconds
//                                   interval: 0..60000 (def: 300 seconds). 
//                                   0 (zero) will disable self-trans.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK        - Success.
//     GWRC_ERROR     - Invalid input parameter
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmSetParameters(
                                         struct hSession *pSession,
                                         unsigned char   gw_max_connection_attempt,
                                         unsigned char   gw_max_send_req_attempt,     
                                         unsigned short  gw_response_timeout,
                                         unsigned short  gw_alive_freq
                                        )
{
   //check validity of parameters:
   if (gw_max_connection_attempt < 1 || gw_max_connection_attempt > 60) return GWRC_ERROR;
   if (gw_max_send_req_attempt < 1 || gw_max_send_req_attempt > 60) return GWRC_ERROR;
   if (gw_response_timeout < 5 || gw_response_timeout > 60) return GWRC_ERROR;
   if (gw_alive_freq > 60000) return GWRC_ERROR;
   if (!pSession) return GWRC_ERROR;

   //set parameters:
   pSession->max_connection_attempt = gw_max_connection_attempt;
   pSession->max_send_req_attempt   = gw_max_send_req_attempt;
   pSession->response_timeout       = gw_response_timeout*1000;
   pSession->alive_freq             = gw_alive_freq / (TIMER_TICK_FRQ / 1000);
   if (gw_alive_freq > 0 && pSession->alive_freq == 0) {
      //okay. alive is set, but the calculation boils downto disabled.
      //set minimum value.
      pSession->alive_freq = TIMER_TICK_FRQ / 1000;
   }
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Changes the key used to Encrypt the data send to, and Decrypt the data
// received from, the Gateway.
// If the key only contains 0 (zero)'s encryption is disabled.
//
// Input:
//      pSession      - Pointer to the session.
//      key           - The 128 bit long encryption key.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK        - Success.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmSetEncryptionKey(struct hSession *pSession,
                                            unsigned char   key[16])
{
   // Validate
   if (!pSession) return GWRC_ERROR;

   // Valid encryption key?
   if (0 == key[0]  && 0 == key[1]  && 0 == key[2]  && 0 == key[3]  &&
       0 == key[4]  && 0 == key[5]  && 0 == key[6]  && 0 == key[7]  &&
       0 == key[8]  && 0 == key[9]  && 0 == key[10] && 0 == key[11] &&
       0 == key[12] && 0 == key[13] && 0 == key[14] && 0 == key[15]) {
      // Disable encryption
      pSession->crypt_key_set = 0;
   } else {
      // Enable encryption
      pSession->crypt_key_set = 1;
   }

   // Set key
   memcpy(pSession->crypt_key, key, 16);

   // Completed
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Enable/disable compression of data send to/from the gateway.
//
// Input:
//      pSession      - Pointer to the session.
//      enable        - 0 = disable compression, <> 0 = Enable compression.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK        - Success.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmEnableCompression(
                                             struct hSession *pSession,
                                             unsigned char   enable
                                            )
{
   // Validate
   if (!pSession) return GWRC_ERROR;

   // Set compression enable/disable
   pSession->comp_enable = (enable==0) ? 0 : 1;

   // Completed
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Enable/disable the use of extended frames, and set maximum allowed
// frame size.
//
// Input:
//      pSession      - Pointer to the session.
//      size          - 0          = disable extended frames
//                      512 - 4096 = Max. allowed size of extended frames.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK        - Success.
//     GWRC_ERROR     - Invalid session
//     GWRC_INV_LEN   - Invalid size.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmEnableExtended(
                                          struct hSession *pSession,
                                          unsigned short  size
                                         )
{
   // Validate
   if (!pSession) return GWRC_ERROR;
   if (size && (size < 512 || size > GW_MAXDATASIZE_EXT)) return GWRC_INV_LEN;

   // Set compression enable/disable
   pSession->ext_maxsize = size;

   // Completed
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Query the maximum frame size allowed by the gateway.
//
// Input:
//      pSession      - Pointer to the session.
//
// Output:
//      None.
//
// Returns:
//     The maximum allowed frame size. (512 - 4096)
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmGetExtended(struct hSession *pSession)
{
   // Validate
   if (!pSession) return GW_MAXDATASIZE_STD;
   if (!pSession->gw_connection || pSession->gw_logon_in_progress) return GW_MAXDATASIZE_STD;

   // Completed
   return pSession->gw_extended_maxsize;
}

/////////////////////////////////////////////////////////////////////////////
// Close the gateway communication.
//
// Input:
//      pSession     - Pointer to the session.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK       - Gateway communication is closed.
//     GWRC_NOT_OPEN - Gateway communication is not open.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmClose(struct hSession *pSession)
{
   // Validate
   if (!pSession) return GWRC_ERROR;
   if (!pSession->opened) return GWRC_NOT_OPEN;

   pSession->opened = 0;

   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Query this nodes node_id.
// This function is used especially when a dynamic node-id is requested.
//
// Input:
//      pSession      - Pointer to the session.
//
// Output:
//      node_id       - Our NodeId.
//
// Returns:
//      GWRC_OK       - Success.
//      GWRC_NOT_OPEN - Gateway communication is not open.
//      GWRC_NOT_CON  - Not connected to a Gateway.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmGetMyNodeID(struct hSession *pSession,
                                       unsigned long   *node_id)
{
   // Validate
   if (!pSession) {
      if (node_id) *node_id = 0;
      return GWRC_ERROR;
   }
   if (!pSession->opened) {
      if (node_id) *node_id = 0;
      return GWRC_NOT_OPEN;
   }
   if (!pSession->gw_connection || pSession->gw_logon_in_progress) {
      if (node_id) *node_id = 0;
      return GWRC_NOT_CON;
   }
   if (node_id) *node_id = pSession->our_node_id;
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Query whether a connection to the gateway is present.
//
// Input:
//      pSession     - Pointer to the session.
//
// Output:
//      None.
//
// Returns:
//      GWRC_OK      - Connected to a Gateway.
//      GWRC_NOT_CON - Not connected to a Gateway.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmIsConnected(struct hSession *pSession)
{
   if (!pSession) return GWRC_NOT_CON;
   return pSession->gw_connection && !pSession->gw_logon_in_progress ? GWRC_OK : GWRC_NOT_CON;
}

/////////////////////////////////////////////////////////////////////////////
// Makes a transaction (request/response) with the specific dst_node via
// the gateway.
//
// Input:
//      pSession         - Pointer to the session.
//      channel          - The channel used for the transaction
//                         interval: 0..7.
//      dst_node         - The NodeID of the destination.
//      cmd              - The command code of the request.
//      datatx           - Pointer to a buffer that contains the data to send.
//      datatx_len       - The length of the data to send. 
//                         interval: 0..GW_MAXDATASIZE.
//      datarx           - Pointer to a buffer where the received date is stored.
//
// Output:
//      rc               - The return code from the destenation.
//      datarx_len       - The length of the received data. (Max. GW_MAXDATASIZE)
//
// Returns:
//      GWRC_OK          - Success.
//      GWRC_NOT_OPEN    - Connection is not opened.
//      GWRC_NOT_CON     - Not connected to a Gateway.
//      GWRC_INV_LEN     - Invalid data length.
//      GWRC_INV_PARM    - Invalid channel.
//      GWRC_DST_UNREACH - The destination was not connected to Gateway
//      GWRC_TOO_LARGE   - Data length is too large for the destination
//      GWRC_TOO_MANY    - There are too many pending messages for the destination
//      GWRC_ERROR       - General error.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmTransaction(
                                       struct hSession *pSession,
                                       unsigned char   channel,
                                       unsigned long   dst_node,
                                       unsigned short  cmd,
                                       const void      *datatx,
                                       unsigned short  datatx_len,
                                       unsigned short  *rc,
                                       void            *datarx,
                                       unsigned short  *datarx_len
                                      )
{
   // Var
   struct request_info  req;
   struct racp2_packet* rp;
   unsigned long        tno;
   unsigned short       ret, j, size;

   // Validate
   if (!pSession) return GWRC_ERROR;
   if (!pSession->opened) return GWRC_NOT_OPEN;
   if (!pSession->gw_connection || pSession->gw_logon_in_progress) {
      return GWRC_NOT_CON;
   }
   if (!pSession->gw_extended_maxsize) {
      if (datatx_len > GW_MAXDATASIZE_STD) {
         return GWRC_INV_LEN;
      }
   } else {
      if (datatx_len > pSession->gw_extended_maxsize) {
         return GWRC_INV_LEN;
      }
   }
   if (channel > 7) {
      return GWRC_INV_PARM;
   }

   ENTER_CRIT;
   tno = pSession->next_tx_tno++;
   EXIT_CRIT;

   for (j = 1; j <= MAX_SEND_REQ_RETRIES; j++) {  
      debug(("[0x%p]gwmTransaction: Attempt %d, TNO=%lu\n", pSession, j, tno));

      //fill out the outstanding request structure:
      req.tno  = tno;
      req.data = datarx;
      SEM_CREATE(req.rxsignal, 0);
      ENTER_CRIT;
      qInsertTail(&pSession->out_requests, (qnode*)&req);
      EXIT_CRIT;
   
      SEM_WAIT(pSession->send_mutex);

      // Select frame
      if (datatx_len > GW_MAXDATASIZE_STD) {
         rp                 =        (struct racp2_packet*)&pSession->req_packet_ext;
      } else {
         rp                 =        &pSession->req_packet_std;
      }

      //fill out the packet to send:
      rp->hdr.src_id        =        pSession->our_node_id;
      rp->hdr.dst_id        =        dst_node;
      rp->hdr.crc           =        0;
      rp->hdr.TNO           =        req.tno;
      rp->hdr.channel       =        channel;
      rp->hdr.encrypt       =        0;
      rp->hdr.compress      =        0;
      rp->hdr.gw_response   =        0;
      rp->hdr.cmd_rc        =        cmd & ~0x8000;  //clear bit 15
      SET_LENGTH(rp, datatx_len);
      if (datatx_len) memcpy(FRAMEDATA(rp), datatx, datatx_len);
      rp->hdr.crc           =        gw_calc_crc16data(0, (unsigned char*)&rp->hdr, HEADLENGTH(rp) + DATALENGTH(rp));
      debug(("[0x%p]gwmTransaction. TNO=%lu crc=0x%x\n", pSession, req.tno, rp->hdr.crc));

      // Compress packet?
      if (pSession->gw_apply_compress && 39 < datatx_len) {
         // Thread safety
         SEM_WAIT(pSession->comp_mutex);
            // Compress data
            size = compress_block((const byte*)datatx, (byte*)pSession->comp_buf, datatx_len);
            // Is data compressed enough?
            if (size < (datatx_len - 3)) {
               // Use compressed data
               memcpy(FRAMEDATA(rp), pSession->comp_buf, size);
               rp->hdr.compress = 1;
               SET_LENGTH(rp, size);
            }
         // Thread safety
         SEM_SIGNAL(pSession->comp_mutex);
      }
      // Set size to send
      size = DATALENGTH(rp);

      // Encrypt packet?
      if (pSession->gw_apply_encrypt && 0 < DATALENGTH(rp)) {
         // Calculate number of blocks
         size = DATALENGTH(rp) / 16;
         if (DATALENGTH(rp) % 16 > 0) size++;

         // Thread safety
         SEM_WAIT(pSession->comp_mutex);
            // Encrypt date
            rijndaelEncryptBuffer128(pSession->crypt_key, FRAMEDATA(rp), (unsigned char*)pSession->comp_buf, size);
            // Use encrypted data
            size *= 16;
            memcpy(FRAMEDATA(rp), pSession->comp_buf, size);
            rp->hdr.encrypt = 1;
         // Thread safety
         SEM_SIGNAL(pSession->comp_mutex);
      } else {
         size = DATALENGTH(rp);
      }

      ret = gwldSend(pSession->conn, (void*)&rp->hdr, HEADLENGTH(rp) + size);

      SEM_SIGNAL(pSession->send_mutex);

      if (ret) {
         //error!!!
         debug(("[0x%p]gwldSend failed. rc=%d. Closing connection\n", pSession,ret));
         ENTER_CRIT;
         qSelRemove(&pSession->out_requests, (qnode*)&req);
         EXIT_CRIT;
         SEM_CLOSE(req.rxsignal);
         gwldDisconnect(pSession->conn);
         pSession->gw_connection = 0;
         return GWRC_ERROR;
      }

      //wait for answer to arrive:
      if (SEM_WAIT_T(req.rxsignal, RESPONSE_TIMEOUT)) {
         debug(("[0x%p]gwmTransaction. Timeout waiting for response. TNO=%lu!\n", pSession, req.tno));
         //timeout:
         ENTER_CRIT;
         qSelRemove(&pSession->out_requests, (qnode*)&req);
         EXIT_CRIT;
         SEM_CLOSE(req.rxsignal);
      } else {
         debug(("[0x%p]gwmTransaction. Response has arrived. TNO=%lu!\n", pSession, req.tno));
         if (channel != 7) pSession->userTransactionsOK = 1;
   
         //response has arrived:
         ENTER_CRIT;
         qSelRemove(&pSession->out_requests, (qnode*)&req);
         EXIT_CRIT;
         SEM_CLOSE(req.rxsignal);
   
         *rc = req.rc & ~0x8000;
         if (datarx_len) *datarx_len = req.data_len;
         if (req.gw_response) {
            return *rc == 1 ? GWRC_DST_UNREACH : (*rc == 4 ? GWRC_TOO_MANY : GWRC_TOO_LARGE);
         }
         return GWRC_OK;
      }
   }
   pSession->forceSelfTransaction = 1;
   return GWRC_ERROR;
}

/////////////////////////////////////////////////////////////////////////////
// Receive a request from a foreign node via the gateway.
// Will block until a request is available.
//
// Input:
//      pSession         - Pointer to the session.
//      channel          - The channel to where the request is coming.
//                         interval: 0..7.
//      data             - Pointer to a buffer where the received date is stored.
//
// Output:
//      src_node         - The NodeID for the source of the request.
//      TNO              - The TNO of the request.
//      cmd              - The command requested by source
//      data_len         - The length of the received data. (Max. GW_MAXDATASIZE)
//
// Returns:
//      GWRC_OK          - Success.
//      GWRC_NOT_OPEN    - Connection is not opened.
//      GWRC_NOT_CON     - Not connected to a Gateway.
//      GWRC_ERROR       - General error.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmReceiveRequest(
                                          struct hSession *pSession,
                                          unsigned char   channel,
                                          unsigned long   *src_node,
                                          unsigned long   *TNO,
                                          unsigned short  *cmd,
                                          void            *data,
                                          unsigned short  *data_len
                                         )
{
   // Var
   struct racp2_packet* rp;

   // Validate
   if (!pSession) return GWRC_ERROR;
   if (!pSession->opened) return GWRC_NOT_OPEN;
   if (channel > 7) return GWRC_INV_PARM;

   //wait for data:
   SEM_WAIT(pSession->in_requests_notempty[channel]);

   //data is here!
   ENTER_CRIT;
   rp = (struct racp2_packet*)qRemoveHead(&pSession->in_requests[channel]);
   EXIT_CRIT;
   if (!rp) {
      //What! - this should never happen..
      return GWRC_ERROR;
   }

   debug(("[0x%p]gwmReceiveRequest. Request received. TNO=%lu!\n", pSession, rp->hdr.TNO));

   //we got the data:
   *src_node = rp->hdr.dst_id;     //dirty. dst_id actually contains the src_id!!
   *TNO      = rp->hdr.TNO;
   *cmd      = rp->hdr.cmd_rc;
   *data_len = DATALENGTH(rp);
   memcpy(data, FRAMEDATA(rp), *data_len);
   freePacket(pSession, rp);
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Receive a request from a foreign node via the gateway.
// Will block until a request is available or timeout expires.
//
// Input:
//      pSession         - Pointer to the session.
//      channel          - The channel to where the request is comming.
//                         interval: 0..7.
//      data             - Pointer to a buffer where the received date is stored.
//
// Output:
//      src_node         - The NodeID for the source of the request.
//      TNO              - The TNO of the request.
//      cmd              - The command requested by source
//      data_len         - The length of the received data. (Max. GW_MAXDATASIZE)
//
// Returns:
//      GWRC_OK          - Success.
//      GWRC_TIMEOUT     - Timeout.
//      GWRC_NOT_OPEN    - Connection is not opened.
//      GWRC_NOT_CON     - Not connected to a Gateway.
//      GWRC_ERROR       - General error.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmReceiveRequestT(
                                           struct hSession *pSession,
                                           unsigned char   channel,
                                           unsigned long   *src_node,
                                           unsigned long   *TNO,
                                           unsigned short  *cmd,
                                           void            *data,
                                           unsigned short  *data_len,
                                           unsigned int    timeout
                                          )
{
   // Var
   struct racp2_packet* rp;

   // Validate
   if (!pSession) return GWRC_ERROR;
   if (!pSession->opened) return GWRC_NOT_OPEN;
   if (channel > 7) return GWRC_INV_PARM;

   //wait for data:
   if (SEM_WAIT_T(pSession->in_requests_notempty[channel], timeout)) {
      debug(("[0x%p]gwmReceiveRequestT: Timeout waiting for request on ch=%d\n", pSession, channel));
      return GWRC_TIMEOUT;
   }

   //data is here!
   ENTER_CRIT;
   rp = (struct racp2_packet*)qRemoveHead(&pSession->in_requests[channel]);
   EXIT_CRIT;
   if (!rp) {
      //What! - this should never happen..
      return GWRC_ERROR;
   }

   debug(("[0x%p]gwmReceiveRequestT. Request received. TNO=%lu!\n", pSession, rp->hdr.TNO));

   //we got the data:
   *src_node = rp->hdr.dst_id;     //dirty. dst_id actually contains the src_id!!
   *TNO      = rp->hdr.TNO;
   *cmd      = rp->hdr.cmd_rc;
   *data_len = DATALENGTH(rp);
   memcpy(data, FRAMEDATA(rp), *data_len);
   freePacket(pSession, rp);
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Send a response back to a node that initiated the transaction.
//
// Input:
//      pSession         - Pointer to the session.
//      dst_node         - The NodeID of the destination.
//      TNO              - The TNO of the received request / transaction.
//      rc               - The Return code of the received request / transaction.
//      data             - Pointer to a buffer that contains the data to send.
//      data_len         - The length of the data to send. 
//
// Output:
//      None.
//
// Returns:
//      GWRC_OK          - Success.
//      GWRC_NOT_OPEN    - Connection is not opened.
//      GWRC_NOT_CON     - Not connected to a Gateway.
//      GWRC_ERROR       - General error.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmSendResponse(
                                        struct hSession *pSession,
                                        unsigned long   dst_node,
                                        unsigned long   TNO,
                                        unsigned short  rc,
                                        const void      *data,
                                        unsigned short  data_len
                                       )
{
   // Var
   struct racp2_packet* rp = 0;
   struct racp2_packet* ptail;
   struct client_info*  ci;
   unsigned short       ret, size;
   int                  found, type;
   debug(("[0x%p]gwmSendResponse. Response submitted. TNO=%lu!\n", pSession, TNO));

   // Validate
   if (!pSession) return GWRC_ERROR;
   if (!pSession->opened) return GWRC_NOT_OPEN;
   if (!pSession->gw_connection) return GWRC_NOT_CON;

   //we need to cache the response:
   //find client in client_list:
   found = 0;
   ENTER_CRIT;
   ci = (struct client_info*)qPeekHead(&pSession->client_list);
   while (ci) {
      if (dst_node == ci->node_id) {
         debug(("[0x%p]gwmSendResponse. Client %lu found in client list.\n", pSession, dst_node));

         //found!
         //move the client_info to the front of the head (MRU)
         qSelRemove(&pSession->client_list, (qnode*)ci);
         qInsertHead(&pSession->client_list, (qnode*)ci);
         found = 1;
         break;
      }
      ci = (struct client_info*)(ci->qlink.next);
   }
   EXIT_CRIT;
   if (!found) {
      debug(("[0x%p]gwmSendResponse. Client %lu NOT found. Response to unknown Request.!\n", pSession, dst_node));
      //this is a weird error! - should not occur...
      return GWRC_ERROR;
   }

   // Select frame type
   if (!pSession->gw_extended_maxsize) {
      if (data_len > GW_MAXDATASIZE_STD) data_len = GW_MAXDATASIZE_STD;
      type = 1;
   } else {
      if (data_len > pSession->gw_extended_maxsize) data_len = pSession->gw_extended_maxsize;
      type = (data_len > GW_MAXDATASIZE_STD) ? 2 : 1;
   }

   ENTER_CRIT;
   ptail = (struct racp2_packet*)qPeekTail(&ci->cacheq); 
   if (qSize(&ci->cacheq) >= CLIENT_CACHE_SIZE) {
      debug(("[0x%p]gwmSendResponse. Max size for cache. Grabbing oldest entry\n", pSession));
      //we have reached max size - remove oldest cache entry:
      rp = (struct racp2_packet*)qSelRemove(&ci->cacheq, qPeekTail(&ci->cacheq));
   }
   EXIT_CRIT;
   if (rp) {
      // Check frame type
      if (rp->type != type) {
         // Wrong type
         freePacket(pSession, rp);
         rp = NULL;
      }
   }
   if (!rp) {
      debug(("[0x%p]gwmSendResponse. Allocating new cache entry\n", pSession));
      rp = allocPacket(pSession, type);
      if (!rp) {
         //trouble!
         return GWRC_ERROR;
      }
   }

   rp->hdr.src_id       =        pSession->our_node_id;
   rp->hdr.dst_id       =        dst_node;
   rp->hdr.crc          =        0;
   rp->hdr.TNO          =        TNO;
   rp->hdr.channel      =        0;      //this may not be the correct channel. The client will not use this info.
   rp->hdr.encrypt      =        0;
   rp->hdr.compress     =        0;
   rp->hdr.cmd_rc       =        (rc&~0x8000)|0x8000; //make sure bit 15 is set.
   SET_LENGTH(rp, data_len);
   memcpy(FRAMEDATA(rp), data, data_len);
   rp->hdr.crc          =        gw_calc_crc16data(0, (unsigned char*)&rp->hdr, HEADLENGTH(rp) + DATALENGTH(rp));
   debug(("[0x%p]gwmSendResponse. TNO=%lu, crc=0x%x\n", pSession, TNO, rp->hdr.crc));

   // Compress packet?
   if (pSession->gw_apply_compress && 39 < data_len) {
      // Thread safety
      SEM_WAIT(pSession->comp_mutex);
         // Compress data
         size = compress_block((const byte*)data, (byte*)pSession->comp_buf, data_len);
         // Is data compressed enough?
         if (size < (data_len - 3)) {
            // Use compressed data
            memcpy(FRAMEDATA(rp), pSession->comp_buf, size);
            rp->hdr.compress = 1;
            SET_LENGTH(rp, size);
         }
      // Thread safety
      SEM_SIGNAL(pSession->comp_mutex);
   }

   // Set size to send
   size = DATALENGTH(rp);

   // Encrypt packet?
   if (pSession->gw_apply_encrypt && 0 < DATALENGTH(rp)) {
      // Calculate number of blocks
      size = DATALENGTH(rp) / 16;
      if (DATALENGTH(rp) % 16 > 0) size++;

      // Thread safety
      SEM_WAIT(pSession->comp_mutex);
         // Encrypt date
         rijndaelEncryptBuffer128(pSession->crypt_key, FRAMEDATA(rp), (unsigned char*)pSession->comp_buf, size);
         // Use encrypted data
         size *= 16;
         memcpy(FRAMEDATA(rp), pSession->comp_buf, size);
         rp->hdr.encrypt = 1;
      // Thread safety
      SEM_SIGNAL(pSession->comp_mutex);
   } else {
      size = DATALENGTH(rp);
   }

   //then we send out the data:
   ret = gwldSend(pSession->conn, (void*)&rp->hdr, HEADLENGTH(rp) + size);
   if (ret) {
      //error!!!
      debug(("[0x%p]gwldSend failed. rc=%d. Closing connection\n", pSession, ret));
      gwldDisconnect(pSession->conn);
      pSession->gw_connection = 0;
   }

   ENTER_CRIT;
   qInsertHead(&ci->cacheq, (qnode*)rp);
   if (ptail != (struct racp2_packet*)qPeekTail(&ci->cacheq)) {
      //new tail. update timestamp:
      ci->last_used = pSession->time_tick;
   }
   EXIT_CRIT;

   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Delete a gateway communication session, this function will free the the
// allocated data, making the pointer invalid.
//
// Output:
//      pSession        - the session handle, will be cleared if it was
//                        possible delete the hole session.
// Returns:
//      GWRC_OK          - Success.
//      GWRC_ERROR       - General error.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwmDelete(struct hSession **pSession)
{
#if GW_DEBUG
   // Var
   void* debug;
#endif

   // Validate
   if (!*pSession) return GWRC_ERROR;
#if GW_DEBUG
   debug = *pSession;
#endif

   // Close
   gwmClose((*pSession));
   (*pSession)->initialized=0;

   // Wait for threads
   debug(("[0x%p]gwmDelete. waiting for worker threads to stop.\n", pSession));
   while ((*pSession)->running_threads) {
      gwldDisconnect((*pSession)->conn);
      SLEEP(100);
   }
   (*pSession)->gw_connection = 0;

   // De-allocate session structure
   gwldDestroy(&(*pSession)->conn);
   free(*pSession);
   *pSession = NULL;
   debug(("[0x%p]gwmDelete - session freed\n", debug));

   // Completed
   return GWRC_OK;
}

/////////////////////////////////////////////////////////////////////////////
// This will initialize/open the gateway communication.
// Must be called before any other function in this DLL (except
// gwSetParameters and gwSetEncryptionKey).
//
// Input:
//      node_id   - Our NodeId. Interval: 0..999999. (0 = Dynamic NodeId).
//      ip_addr   - String containing Gateway IP address. (can be symbolic)
//      port      - The Gateway IP port.
//      key       - String containing Gateway key. (Max. 8 characters)
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK       - Success.
//     GWRC_IS_OPEN  - Gateway communication is already open.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwOpen(
                               unsigned long  node_id,
                               const char     *ip_addr,
                               unsigned short port,
                               const char     key[9]
                              )
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmOpen(gwSession, node_id, ip_addr, port, key);
}

/////////////////////////////////////////////////////////////////////////////
// Query whether the gateway communication is open.
//
// Input:
//      None.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK       - Gateway communication is open.
//     GWRC_NOT_OPEN - Gateway communication is not open.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwIsOpened(void)
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmIsOpened(gwSession);
}

/////////////////////////////////////////////////////////////////////////////
// Sets the TLS parameters of a session.
//
// Input:
//      server                     - The path to the PEM file used to verify
//                                   the server certificate.
//      cert                       - The client certificate to present to
//                                   the server. Reserved for future use.
//      key                        - The private key of the client.
//                                   Reserved for future use.
//      password                   - The password of the private key.
//                                   Reserved for future use.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK        - Success.
//     GWRC_ERROR     - Invalid input parameter
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwSetTLSParameters(
                                           char*            server,
                                           char*            cert,
                                           char*            key,
                                           char*            password
                                          )
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmSetTLSParameters(gwSession, server, cert, key, password);
}

/////////////////////////////////////////////////////////////////////////////
// Sets the Advanced parameters.
//
// Input:
//      gw_max_connection_attempt  - max number of connection attempts before GPRS re-connects.
//                                   interval: 1..60. (default: 3) (NOT USED ON PC IMPLEMENTATION)
//      gw_max_send_req_attempt    - max number of send-request attempt before send fails.
//                                   interval: 1..60 (default: 3).
//      gw_response_timeout        - time waiting for response in seconds.
//                                   interval: 5..60 (def: 30 seconds).
//      gw_alive_freq              - frequency for sending self-transactions in seconds
//                                   interval: 0..60000 (def: 300 seconds). 
//                                   0 (zero) will disable self-trans.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK        - Success.
//     GWRC_ERROR     - Invalid input parameter
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwSetParameters(
                                        unsigned char  gw_max_connection_attempt,
                                        unsigned char  gw_max_send_req_attempt,     
                                        unsigned short gw_response_timeout,
                                        unsigned short gw_alive_freq
                                       )
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmSetParameters(gwSession,
                           gw_max_connection_attempt,
                           gw_max_send_req_attempt,
                           gw_response_timeout,
                           gw_alive_freq);
}

/////////////////////////////////////////////////////////////////////////////
// Changes the key used to Encrypt the data send to, and Decrypt the data
// received from, the Gateway.
// If the key only contains 0 (zero)'s encryption is disabled.
//
// Input:
//      key           - The 128 bit long encryption key.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK        - Success.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwSetEncryptionKey(unsigned char key[16])
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmSetEncryptionKey(gwSession, key);
}

/////////////////////////////////////////////////////////////////////////////
// Enable/disable the use of extended frames, and set maximum allowed
// frame size.
//
// Input:
//      size          - 0          = disable extended frames
//                      512 - 4096 = Max. allowed size of extended frames.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK        - Success.
//     GWRC_INV_LEN   - Invalid size.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwEnableExtended(unsigned short size)
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmEnableExtended(gwSession, size);
}

/////////////////////////////////////////////////////////////////////////////
// Query the maximum frame size allowed by the gateway.
//
// Input:
//      None.
//
// Output:
//      None.
//
// Returns:
//     The maximum allowed frame size. (512 - 4096)
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwGetExtended()
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmGetExtended(gwSession);
}

/////////////////////////////////////////////////////////////////////////////
// Close the gateway communication.
//
// Input:
//      None.
//
// Output:
//      None.
//
// Returns:
//     GWRC_OK       - Gateway communication is closed.
//     GWRC_NOT_OPEN - Gateway communication is not open.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwClose(void)
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmClose(gwSession);
}

/////////////////////////////////////////////////////////////////////////////
// Query this nodes node_id.
// This function is used especially when a dynamic node-id is requested.
//
// Input:
//      None.
//
// Output:
//      node_id       - Our NodeId.
//
// Returns:
//      GWRC_OK       - Success.
//      GWRC_NOT_OPEN - Gateway communication is not open.
//      GWRC_NOT_CON  - Not connected to a Gateway.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwGetMyNodeID(unsigned long* node_id)
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmGetMyNodeID(gwSession, node_id);
}

/////////////////////////////////////////////////////////////////////////////
// Query whether a connection to the gateway is present.
//
// Input:
//      None.
//
// Output:
//      None.
//
// Returns:
//      GWRC_OK      - Connected to a Gateway.
//      GWRC_NOT_CON - Not connected to a Gateway.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwIsConnected(void)
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmIsConnected(gwSession);
}

/////////////////////////////////////////////////////////////////////////////
// Makes a transaction (request/response) with the specific dst_node via
// the gateway.
//
// Input:
//      channel          - The channel used for the transaction
//                         interval: 0..7.
//      dst_node         - The NodeID of the destination.
//      cmd              - The command code of the request.
//      datatx           - Pointer to a buffer that contains the data to send.
//      datatx_len       - The length of the data to send. 
//                         interval: 0..GW_MAXDATASIZE.
//      datarx           - Pointer to a buffer where the received date is stored.
//
// Output:
//      rc               - The return code from the destenation.
//      datarx_len       - The length of the received data. (Max. GW_MAXDATASIZE)
//
// Returns:
//      GWRC_OK          - Success.
//      GWRC_NOT_OPEN    - Connection is not opened.
//      GWRC_NOT_CON     - Not connected to a Gateway.
//      GWRC_INV_LEN     - Invalid data length.
//      GWRC_INV_PARM    - Invalid channel.
//      GWRC_DST_UNREACH - The destination was not connected to Gateway
//      GWRC_TOO_LARGE   - Data length is too large for the destination
//      GWRC_TOO_MANY    - There are too many pending messages for the destination
//      GWRC_ERROR       - General error.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwTransaction(
                                      unsigned char  channel,
                                      unsigned long  dst_node,
                                      unsigned short cmd,
                                      const void     *datatx,
                                      unsigned short datatx_len,
                                      unsigned short *rc,
                                      void           *datarx,
                                      unsigned short *datarx_len
                                     )
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmTransaction(gwSession,
                         channel,
                         dst_node,
                         cmd,
                         datatx,
                         datatx_len,
                         rc,
                         datarx,
                         datarx_len);
}

/////////////////////////////////////////////////////////////////////////////
// Receive a request from a foreign node via the gateway.
// Will block until a request is available.
//
// Input:
//      channel          - The channel to where the request is coming.
//                         interval: 0..7.
//      data             - Pointer to a buffer where the received date is stored.
//
// Output:
//      src_node         - The NodeID for the source of the request.
//      TNO              - The TNO of the request.
//      cmd              - The command requested by source
//      data_len         - The length of the received data. (Max. GW_MAXDATASIZE)
//
// Returns:
//      GWRC_OK          - Success.
//      GWRC_NOT_OPEN    - Connection is not opened.
//      GWRC_NOT_CON     - Not connected to a Gateway.
//      GWRC_ERROR       - General error.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwReceiveRequest(
                                         unsigned char  channel,
                                         unsigned long  *src_node,
                                         unsigned long  *TNO,
                                         unsigned short *cmd,
                                         void           *data,
                                         unsigned short *data_len
                                        )
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmReceiveRequest(gwSession, channel, src_node, TNO, cmd, data, data_len);
}

/////////////////////////////////////////////////////////////////////////////
// Receive a request from a foreign node via the gateway.
// Will block until a request is available or timeout expires.
//
// Input:
//      channel          - The channel to where the request is comming.
//                         interval: 0..7.
//      data             - Pointer to a buffer where the received date is stored.
//
// Output:
//      src_node         - The NodeID for the source of the request.
//      TNO              - The TNO of the request.
//      cmd              - The command requested by source
//      data_len         - The length of the received data. (Max. GW_MAXDATASIZE)
//
// Returns:
//      GWRC_OK          - Success.
//      GWRC_TIMEOUT     - Timeout.
//      GWRC_NOT_OPEN    - Connection is not opened.
//      GWRC_NOT_CON     - Not connected to a Gateway.
//      GWRC_ERROR       - General error.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwReceiveRequestT(
                                          unsigned char  channel,
                                          unsigned long  *src_node,
                                          unsigned long  *TNO,
                                          unsigned short *cmd,
                                          void           *data,
                                          unsigned short *data_len,
                                          unsigned int   timeout
                                         )
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmReceiveRequestT(gwSession,
                             channel,
                             src_node,
                             TNO,
                             cmd,
                             data,
                             data_len,
                             timeout);
}

/////////////////////////////////////////////////////////////////////////////
// Send a response back to a node that initiated the transaction.
//
// Input:
//      dst_node         - The NodeID of the destination.
//      TNO              - The TNO of the received request / transaction.
//      rc               - The Return code of the received request / transaction.
//      data             - Pointer to a buffer that contains the data to send.
//      data_len         - The length of the data to send. 
//
// Output:
//      None.
//
// Returns:
//      GWRC_OK          - Success.
//      GWRC_NOT_OPEN    - Connection is not opened.
//      GWRC_NOT_CON     - Not connected to a Gateway.
//      GWRC_ERROR       - General error.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short GWPROTCC gwSendResponse(
                                       unsigned long  dst_node,
                                       unsigned long  TNO,
                                       unsigned short rc,
                                       const void     *data,
                                       unsigned short data_len
                                      )
{
   if (!gwSession) gwmCreate(&gwSession);
   return gwmSendResponse(gwSession, dst_node, TNO, rc, data, data_len);
}


/****************************************************************************
**
**    EOF 
**
****************************************************************************/
