/***************************** C SOURCE FILE *********************************
**
**  Project:    Logic IO RTCU Gateway Protocol
**  Filename:   SECURITY.C
**
******************************************************************************
**
**  This is the functions for security
**
*****************************************************************************/


/*****************************************************************************
**
** MODULES USED
**
*****************************************************************************/
#include <stdio.h>
#include <openssl/bio.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include "security.h"



/*****************************************************************************
**
**    DEFINITIONS AND MACROS
**
*****************************************************************************/


/****************************************************************************
**
**    TYPEDEFS AND STRUCTURES
**
****************************************************************************/


/*****************************************************************************
**
** EXPORTED VARIABLES
**
*****************************************************************************/


/*****************************************************************************
**
** GLOBAL VARIABLES
**
*****************************************************************************/


/*****************************************************************************
**
** LOCAL FUNCTIONS
**
*****************************************************************************/


/*****************************************************************************
**
** EXPORTED FUNCTIONS
**
*****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
// Validate certificate file
//
// Input:
//      filename         - The filename of the certificate to validate.
//
// Output:
//      None.
//
// Returns:
//      0                - Not a valid certificatey.
//      1                - Certificate is valid.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short ValidateCertificate(const char *filename)
{
   // Var
   BIO* bio;
   X509* pem;
   FILE* fp;
   size_t size, len;
   unsigned char* data;

   // Open file
   fp = fopen(filename, "rb");
   if (fp == NULL) return 0;

   // Get file size
   fseek(fp, 0, SEEK_END);
   size = ftell(fp);
   fseek(fp, 0, SEEK_SET);

   // Allocate
   data = (unsigned char*)malloc(size);
   if (data == NULL) {
      fclose(fp);
      return 0;
   }

   // Read contents
   len = fread(data, 1, size, fp);
   if (len != size) {
      fclose(fp);
      free(data);
      return 0;
   }

   // Close file
   fclose(fp);

   // Prepare OpenSSL
   bio = BIO_new(BIO_s_mem());
   if (bio == NULL) {
      free(data);
      return 0;
   }
   len = BIO_write(bio, (const void*)data, size);
   if (len != size)
   {
      BIO_free(bio);
      free(data);
      return 0;
   }

   // Get certificate
   pem = PEM_read_bio_X509(bio, NULL, 0, NULL);
   if (pem == NULL)
   {
      BIO_free(bio);
      free(data);
      return 0;
   }

   // Cleanup
   X509_free(pem);
   BIO_free(bio);
   free(data);

   // Completed
   return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Validate certificate file
//
// Input:
//      cert             - The filename of the certificate to validate.
//      key              - The filename of the private key to validate.
//      pass             - The password of the private key.
//
// Output:
//      None.
//
// Returns:
//      0                - Not a valid certificatey.
//      1                - Certificate is valid.
//
/////////////////////////////////////////////////////////////////////////////
unsigned short ValidateClient(const char* cert, const char* key, const char* pass)
{
   // Var
   SSL_CTX* tls_ctx;

   // Create TLS context
   tls_ctx = SSL_CTX_new(TLS_method());
   if (tls_ctx == NULL) return 0;

   // Set password
   SSL_CTX_set_default_passwd_cb_userdata(tls_ctx, pass);

   // Set certificate
   if (!SSL_CTX_use_certificate_file(tls_ctx, cert, SSL_FILETYPE_PEM)
       || !SSL_CTX_use_PrivateKey_file(tls_ctx, key, SSL_FILETYPE_PEM)
       || !SSL_CTX_check_private_key(tls_ctx))
   {
      // Clean up
      SSL_CTX_free(tls_ctx);

      // Completed
      return 0;
   }

   // Clean up
   SSL_CTX_free(tls_ctx);

   // Completed
   return 1;
}


/****************************************************************************
**
**    EOF
**
****************************************************************************/
